/* This file is part of the Imote2 Codeloader.
 *
 * The Imote2 Codeloader is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Imote2 Codeloader is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Imote2 Codeloader. If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * Client.java
 *
 * @author Markus Engel
 * Copyright (c) 2009,2010 Networked Systems Group, University of Kaiserslautern
 */
package imote;

import ch.ntb.usb.USBException;
import imote.communication.ImoteException;
import java.io.File;
import java.io.IOException;
import java.rmi.ConnectException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import imote.service.ImoteInterface;
import imote.service.ServiceInterface;
import java.io.RandomAccessFile;
import java.util.List;
import org.jvcompress.lzo.MiniLZO;
import org.jvcompress.util.MInt;

/**
 * RMI Client Role
 */
public class Client implements Role {
	public void start(Configuration config) {
		/* entry point: connect to RMI Server, lookup Service and process
		 * configuration
		 */
		try {
			Registry registry = LocateRegistry.getRegistry(config.getConnect());
			ServiceInterface iface = (ServiceInterface) registry.lookup(Server.SERVICE);
			System.out.println("Connected Service at " + config.getConnect());

			processConfig(config, iface);
		} catch (ConnectException e) {
			System.out.println("Could not connect to Service at "
				+ config.getConnect() + ": " + e.toString());
		} catch (NotBoundException e) {
			System.out.println("Could not bind Service \"" + Server.SERVICE
				+ "\" at " + config.getConnect() + ": " + e.toString());
		} catch (RemoteException e) {
			System.out.println("RemoteException occured while connecting: "
				+ e.toString());
		}
	}

	/** process configuration on a given service interface
	 *
	 * @param config configuration
	 * @param iface service interface
	 */
	public static void processConfig(Configuration config,
		ServiceInterface iface) {
		try {
			if (config.getListImotes()) {
				/* just list imotes */
				listImotes(config, iface);
			} else {
				/* process upload requests */
				doUpload(config, iface);
			}
		} catch (ImoteException iex) {
			System.err.println("Exception on Imote occured: " + iex.toString());
		} catch (USBException usbe) {
			System.err.println("Exception in USB Communication occured: "
				+ usbe.toString());
		} catch (IOException ioe) {
			System.err.println("IOException occured: " + ioe.toString());
		} catch (Exception e) {
			System.err.println("Exception occured: " + e.toString());
			e.printStackTrace();
		} finally {
			try {
				iface.closeImote();
			} catch (Exception e) {
			}
		}
	}

	/** list all imotes
	 *
	 * @param config configuration
	 * @param iface service interface
	 * @throws USBException whenever there goes something wrong concerning USB
	 * @throws RemoteException RMI Exceptions
	 */
	private static void listImotes(Configuration config, ServiceInterface iface)
		throws USBException, RemoteException {
		List<ImoteDescription> imotes =
			iface.getConnectedImotes(config.getVendorId(), config.getProductId());
		System.out.println("Connected Imotes:");
		for (ImoteDescription imote : imotes) {
			ImoteSerial ser = imote.getSerial();
			System.out.println("\tSerNo: " + ser.serialNumber +
				(ser.nodeId != null ? " Node ID: " + ser.nodeId : "") +
				" Bootloader: " + imote.getVersion());
		}
	}

	/** process upload requests
	 *
	 * @param config configuration
	 * @param iface service interface
	 * @throws ImoteException whenever something with the mote fails
	 * @throws USBException whenever there goes something wrong concerning USB
	 * @throws IOException when a upload file cannot be read
	 * @throws Exception any other errors
	 */
	private static void doUpload(Configuration config, ServiceInterface iface)
		throws ImoteException, USBException, IOException, Exception {
		ImoteInterface imote = null;
		int i;

		System.out.print("Waiting for device");
		System.out.flush();
		i = 1;
		/* try to get the imote which is described by the configuration through
		 * the service interface*/
		String serNo;
		if (config.hasSerialNumber()) {
			serNo = config.getSerialNumber();
		} else {
			serNo = null;
		}
		Long nodeId;
		if (config.hasNodeId()) {
			nodeId = config.getNodeId();
		} else {
			nodeId = null;
		}
		while (true) {
			if (i == 10) {
				System.out.print(".");
				System.out.flush();
				i = 1;
			} else {
				++i;
			}
			try {
				imote = iface.getImote(config.getVendorId(),
					config.getProductId(), serNo, nodeId);
			} catch (USBException e) {
				System.err.println("Exception Occured: " + e
					+ ", trying again...");
			}
			if (imote != null) {
				break;
			}
			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
			}
		}

		/* print some info */
		System.out.println("\nFound device, Bootloader Version "
			+ imote.getDeviceVersion());

		ImoteSerial ser = imote.getSerial();
		if (ser.serialNumber.equals("") || ser.serialNumber.equals("ffffffff")) {
			System.out.println("Warning: serialnumber not set on device");
		} else {
			System.out.println("Serialnumber: " + ser.serialNumber);
		}

		if (ser.nodeId == null || ser.nodeId.equals(0xffffffffl)) {
			System.out.println("Warning: node id not set on device");
		} else {
			System.out.println("NodeId: " + ser.nodeId);
		}

		/* sanity check */
		if (config.hasRamExecutionRequest()
			&& imote.getBcdDeviceVersion() < 0x105) {
			System.err.println("Execution from memory requested, but "
				+ "bootloader is too old (at least 1.5 needed)");
			return;
		}

		/* begin protocol */
		System.out.println("Resetting device (State IDLE)");
		imote.reset();

		if (config.getInvalidate()) {
			/* invalidate default image location */
			System.out.println("Invalidating default image location");
			if (imote.getBcdDeviceVersion() >= 0x102) {
				imote.address(Configuration.UploadRequest.DEFAULT_TARGET_ADDRESS,
					0);
				imote.erase();
				imote.uploadFinish();
			} else {
				imote.address(Configuration.UploadRequest.DEFAULT_TARGET_ADDRESS,
					64);
				imote.erase();
				imote.upload(new byte[64], 0, 64, false);
				imote.uploadFinish();
			}
		}

		/* process all requests */
		for (Configuration.UploadRequest request : config.getRequests()) {
			if (request.getExecuteRam() == true) {
				/* skip that one (done at the end) */
				continue;
			}
			System.out.println("Uploading " + request.getFileName() + " @ "
				+ String.format("0x%08x", request.getTargetAddress()));
			uploadFile(request, imote, config);
		}

		if (config.hasRamExecutionRequest()) {
			/* now do the execute-in-ram request */
			Configuration.UploadRequest request = config.getRamExecutionRequest();
			System.out.println("Uploading " + request.getFileName()
				+ " into ram");
			uploadFile(request, imote, config);

			System.out.println("Executing Image from Memory (State EXEC)");
			imote.executeRam();
			return;
		}

		if (config.getReboot()) {
			System.out.println("Rebooting device (State REBOOT)");
			imote.reboot();
		}
	}

	/** process one request (upload file)
	 *
	 * @param request request to be processed
	 * @param imote imote to upload to
	 * @param config configuration
	 * @throws ImoteException whenever something with the mote fails
	 * @throws USBException whenever there goes something wrong concerning USB
	 * @throws IOException when a upload file cannot be read
	 * @throws Exception any other errors
	 */
	private static void uploadFile(Configuration.UploadRequest request,
		ImoteInterface imote, Configuration config)
		throws ImoteException, USBException, IOException, Exception {
		/* sanity check: that won't be accepted by bootloader either */
		if (request.getTargetAddress() < 0x00040000l) {
			System.err.printf("Skipping file, Target Address 0x%08x "
				+ "does not match the address constraints\n",
				request.getTargetAddress());
			return;
		}

		/* some other sanity checks */
		File file = new File(request.getFileName());
		if (!file.exists()) {
			System.err.println("File " + file.toString() + " does not exist");
			return;
		}
		if (!file.canRead()) {
			System.err.println("File " + file.toString() + " is not readable");
			return;
		}

		/* sanity check: that won't be accepted by bootloader either */
		if (request.getTargetAddress() + file.length() > (32 << 20)
			&& request.getExecuteRam() == false) {
			System.err.printf("Skipping file, Target Address and Length "
				+ "out of bounds");
			return;
		}

		RandomAccessFile input = new RandomAccessFile(file, "r");
		try {
			if (request.isExecutable()) {
				/* target address tells us that this image will be executed
				 * so we try to read the image to do some sanity checks
				 */
				Image img = new Image(input);
				if (!img.isBootable()) {
					/* image not recognized */
					System.err.println("ERROR: you are about to upload a very "
						+ "old image or \n\tsome file which is not executable "
						+ "all\n\t(this could also be an ELF file, try "
						+ "component.elf.out)");
					if (!config.getForceUpload()) {
						throw new Exception("Won't upload that file");
					}
				}
				if (img.hasVersion()) {
					/* some information */
					Image.Version ver = img.getVersion();
					System.out.printf("INFO: Image header version %d.%d " +
							"detected\n", ver.getMajor(), ver.getMinor());
					if (img.hasDescription()) {
						System.out.printf("\tImage description: %s\n",
							img.getDescription());
					}
				}
			}

			/* follow the protocol... */
			System.out.println("Sending Address and Length (State ADDR)");
			if (request.getExecuteRam() == false) {
				imote.address(request.getTargetAddress(), file.length());

				System.out.println("Erasing... (State ERASE)");
				imote.erase();
			} else {
				imote.address(Configuration.UploadRequest.RAM_TARGET_ADDRESS,
					file.length());
				System.out.println("Skipping Erase (Memory Execution)");
				imote.erase();
			}

			long uploadStartTime = new java.util.Date().getTime();
			System.out.print("Uploading (State UPLOADING)");

			/* compute of some parameters according to bootloader
			 * version (=> wether bootloader knows about compression) and so on
			 */
			UploadParameters p = new UploadParameters(
				config.getCompress(),
				imote.getBcdDeviceVersion() >= 0x200,
				request.getExecuteRam());

			/* number of blocks read */
			int blocks = 0;
			/* number of blocks which could be compressed */
			int compressedBlocks = 0;
			/* blockSize, determined by maximal transfer size, compression and
			 * alignments */
			byte [] rdBuf = p.getBuffer();
			/* number of bytes actually sent */
			long rawLength = 0;
			/* number of packets sent */
			int packets = 0;

			byte [] rdCompressed = p.getCompressBuffer();

			for (int i = 0; i < file.length(); i += p.getBlockSize()) {
				long percentage = i * 100L / file.length();
				long lengthPercentage = 100 - (i == 0 ? 0 : rawLength * 100L / i);
				System.out.printf("\rUploading (State UPLOADING) [%3d%%] [%3d%%]",
					percentage, lengthPercentage);
				System.out.flush();

				/* read into buffer */
				int read = input.read(rdBuf);
				blocks++;
				if (p.getCompress()) {
					/* try to compress */
					int c = compress(rdBuf, rdCompressed, read);
					if (c <= 0 || c > p.getMaxPacketSize()) {
						/* compressed data, buf it was not really shorter,
						 * thus send the read data in max packets chunks */
						for (int j = 0; j < read; j += p.getChunkSize()) {
							int sz = read - j <  p.getChunkSize() ?
								read - j :  p.getChunkSize();
							imote.upload(rdBuf, j, sz, false);
							packets++;
						}
						rawLength += read;
					} else {
						/* data could be compressed, send them */
						imote.upload(rdCompressed, 0, c, true);
						packets++;
						rawLength += c;
						compressedBlocks++;
					}
				} else {
					/* no compression wanted, just send the data */
					imote.upload(rdBuf, 0, read, false);
					packets++;
					rawLength += read;
				}
			}
			
			long uploadDuration = new java.util.Date().getTime() - uploadStartTime;
			String uploadSpeed;
			if (uploadDuration == 0) {
				uploadSpeed = "?";
			} else {
				uploadSpeed = new Long((file.length() / 1024) * 1000 / uploadDuration).toString();
			}
			System.out.printf("\rUploading (State UPLOADING) [100%%]\n\t" +
				"%s kb/s, %d blocksize, %d%% transfer usage,\n\t" +
				"%d%% block compression ratio, %d%% overall compression\n",
				uploadSpeed, p.getBlockSize(),
				rawLength * 100L / (p.getMaxPacketSize() * packets),
				blocks != 0 ? (compressedBlocks * 100L / blocks) : 0L,
				100 - (rawLength * 100L / file.length()));

			System.out.println("Finishing upload (State IDLE)");
			System.out.flush();
			imote.uploadFinish();
		} finally {
			input.close();
		}
	}

	/** helper to compress data using LZO
	 *
	 * @param in input bytes
	 * @param out output bytes
	 * @param sz input size
	 * @return compressed size or 0 if compression failed
	 */
	private static int compress(byte [] in, byte [] out, int sz) {
		int [] dict = new int[128 * 1024];
		MInt outL = new MInt();

		try {
			int r = MiniLZO.lzo1x_1_compress(in, sz, out, outL, dict);
			if (r != MiniLZO.LZO_E_OK || outL.v >= sz) {
				return 0;
			}

			return outL.v;
		} catch(Exception e) {
			return 0;
		}
	}
}
