/* This file is part of the Imote2 Codeloader.
 *
 * The Imote2 Codeloader is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Imote2 Codeloader is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Imote2 Codeloader. If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * UploadParameters.java
 *
 * @author Markus Engel
 * Copyright (c) 2009,2010 Networked Systems Group, University of Kaiserslautern
 */
package imote;

/** Helper class to calculate upload parameters
 */
class UploadParameters {
	/* expermimental value, must match bootloader */
	/** try to read the MAX_DATA_SIZE_20 times the READ_SIZE_MULT and
	 * compress it into (less or equal) MAX_DATA_SIZE_20 bytes
	 */
	private static final float READ_SIZE_MULT = 1.3f;
	/** max data size for bootloader version 2.0 and higher */
	private static final int MAX_DATA_SIZE_20 =
		8191 - imote.communication.Packet.HEADER_SIZE;
	/** max data size for bootlaoder version 1.0 */
	private static final int MAX_DATA_SIZE_10 = 2048;

	/** packet size */
	private int packetSize;
	/** chunk size */
	private int chunkSize;
	/** block size */
	private int blockSize;
	/** wether to compress */
	private boolean compress;
	/** wether we have a newer bootloader (2.0 and above) */
	private boolean newerBootloader;
	/** wether we should execute in ram */
	private boolean execRam;

	UploadParameters(boolean shouldCompress, boolean newerBootloader,
		boolean execRam) {
		this.compress = shouldCompress && newerBootloader;
		this.newerBootloader = newerBootloader;
		this.execRam = execRam;

		this.calculate();
	}

	private void calculate() {
		if (!this.newerBootloader) {
			/* older bootloader: read and send as much as possible
			 * (2k data + 4 header) */
			this.blockSize = MAX_DATA_SIZE_10;
			this.packetSize = MAX_DATA_SIZE_10;
			this.chunkSize = 0;
		} else {
			/* newer bootloader */
			this.packetSize = MAX_DATA_SIZE_20;
			this.chunkSize = MAX_DATA_SIZE_20;
			if (this.compress) {
				this.blockSize = (int)(READ_SIZE_MULT * MAX_DATA_SIZE_20);
			} else {
				this.blockSize = MAX_DATA_SIZE_20;
			}
			if (this.execRam) {
				/* upload to ram => align on word-basis */
				this.blockSize &= ~(4 - 1);
				this.chunkSize &= ~(4 - 1);
			} else {
				/* upload to flash => align on flash cell size (64)
				 * this is a MUST! */
				this.blockSize &= ~(64 - 1);
				this.chunkSize &= ~(64 - 1);
			}
		}
	}

	byte [] getBuffer() {
		return new byte[this.blockSize];
	}

	byte [] getCompressBuffer() {
		return this.compress ? new byte[2 * this.blockSize] : null;
	}

	int getBlockSize() {
		return this.blockSize;
	}

	int getChunkSize() {
		return this.chunkSize;
	}

	boolean getCompress() {
		return this.compress;
	}

	int getMaxPacketSize() {
		return this.packetSize;
	}
}
